// Billardsimulation: Simulation vllig reibungsfreier, vollelastischer
// Stsse (Eta=1) als animierte, farbige Realtime-Simulation
// (c) 1996, 1997 by Andreas Meile, 8242 Hofen SH
// Erweiterte Version mit optionaler Gravitation (Schwerkraft wie bei
// einem Flipperkasten)
// Ausserdem: Halde fr Ereignisschlange
// Version mit gefllter Przisionsgrafik

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <grx.h>
#include <conio.h>
#include <pc.h>
#include "eing_rout.h"

// Eigene Datentypen

class vektor {  // 2D-vektor
  public:
    float x,      // x-Komponente
          y;      // y-Komponente
};

class bandenwand { // Beschreibung eines Bandenwandstckes
  public:
    vektor b,  // Eckpunkt
           e,  // Einheitsvektor, gerichtet zum nchsten Eckpunkt
           e2; // Transformationsvektor: Entspricht e bei komplexer
               // Betrachtungsweise
    float a,   // Abstand des Koordinatenursprungs zur Bandenwand
          g;   // Querkomponente Schwerkraftbeschleunigungskonstante g
};

class kugel {     // Beschreibung einer Billardkugel
  public:
    vektor p,       // Position zur Zeit t0
           v;       // momentane Geschwindigkeit
    float t0,       // Bezugszeitpunkt t0
          m,        // Masse
          r,        // Radius, physikalischer
          ra_x,     // umgerechnete Radien horizontal
          ra_y;     // umgerechnete Radien vertikal
    int f;          // Farbe

    // Kugel nachrcken
    void bewege (float t);
    // Kugel zeichnen
    void zeichne ();
    // Umlenkung mit Bande berechnen
    void umlenkung_bande (const bandenwand &bw, float t);
    // Umlenkung mit anderer Kugel berechnen
    void umlenkung_stoss (kugel &k2, float t);
    // Zeitpunkt fr Zusammenstoss mit anderer Kugel bestimmen
    int wann_stoss (const kugel &k2, float &t);
    // Zeitpunkt fr Bandenstoss bestimmen
    int wann_bande (const bandenwand &bw, float &t);
    // Zeitpunkt fr Bandenstoss bestimmen, gesamte Bande
    int wann_bande_rundum (const bandenwand b[], float &zp_min);
};

struct stosszeit {  // stosszeiteintrag
  float t;    // Zeitpunkt
  int eind1,  // Index auf 1. Element
      eind2;  // Index auf 2. Element
};

const int n_waende_max = 50;       // Bande darf max. aus 50 Wnde bestehen
const int n_kugeln_max = 100;      // Max. 100 Kugeln im Szenario
const float y_x_verh = 3.0 / 4.0;  // Bildverhltnis
const float unendl = 3.402823E+38; // wirkt praktisch als unendlich ()
const float pi = 4.0 * atan(1.0);  // Konstante 
const int n_st_ereig_max = n_kugeln_max * (n_kugeln_max + 1) + 2;

// Globale Variablen
stosszeit st_ereig[n_st_ereig_max],  // Halde fr Stosszeiten
          sto;
int anz_st = 0,         // Anzahl derzeit mgliche Stsse
    ton_ein = 1,        // Flag, ob mit oder ohne Sound
    n_kugeln = 0,       // Anzahl Kugeln in der Szene (anfnglich noch keine)
    n_waende = 0,       // Anzahl Wnde in der Szene
    ws,                 // hchster Bandenpunkt (Index)
    we,                 // niedrigster Bandenpunkt
    sound_stat = 0,     // Soundstatus: 0 = still, 1 = Kugel-Kugel,
                        // 2 = Kugel-Bande (tieferer Ton) gerade klingend
    ton_aus_sperre = 0, // Flag, dass der Ton nicht sofort wieder
                        // abgeschaltet werden kann: Ein Ton soll immer
                        // mindestens eine Aufbauperiode lang tnen, um ein
                        // Knacken zu vermeiden
    x_min1 = 0,         // linke Begrenzung um Kugeln
    x_max1,             // rechte Begrenzung um Kugeln
    y_min1 = 0,         // obere Begrenzung um Kugeln
    y_max1,             // untere Begrenzung um Kugeln
    
    alter_videomodus,   // fr Videomodussicherung
    treiber_konst = GR_width_height_color_graphics, // Hardware-Konstante fr
    // Grafikaktivierung; falls mit lterem Treiber ohne Explizit-Modusangabe
    // gearbeitet wird, muss vom Benutzer ein anderer Wert eingegeben werden
    x_aufl = 640,  // horizontale Grafikauflsung
    y_aufl,        // vertikale Grafikauflsung
    a_far = 16;    // Anzahl Farben

float verkl_fakt = 1.01,  // Verkleinerungsfaktor; wre dieser 1.0, dann
                          // wrde die Bande direkt den Bildschrimrand
                          // berhren (Steuerung der Grsse am Bildschirm)
      verla_fakt = 1.0, // Verlangsamungsfaktor; erlaubt eine Geschwindig-
                        // keitsrcknahme bei langsameren PC's
      kf,               // Verzerrungsfaktor Pixelhhe/Pixelbreite
      fx,               // Umrechnungsfaktor horizontal
      fy,               // Umrechnungsfaktor vertikal
      x_ba,             // Verschiebungsbetrag horizontal
      y_ba,             // Verschiebungsbetrag vertikal
      akt_zeit = -500;  // aktueller Zeitpunkt

vektor g = {0.0, 0.0};  // Beschleunugung, vorgabemssig keine Schwerkraft
                        // => Weltall-Verhltnisse

bandenwand bande[n_waende_max]; // Bandenwand der Szene
kugel b_kugel[n_kugeln_max];    // Kugeln der Szene

GrContext *leere_bande, // Leere Bande
          *szene;       // Szene im Hintergrund

// Quadrat eines Wertes bestimmen
float quad(float s) {
  return s * s;
}

void stoss_einfuegen (const stosszeit &s) {
  anz_st++;
  register int t = anz_st;
  do {
    register int t2 = t >> 1;
  if(st_ereig[t2].t <= s.t) break;
    st_ereig[t] = st_ereig[t2];   // kleiner als die Eltern => nachrcken
    t = t2;
  } while(1);
  st_ereig[t] = s;
}

void loesche_ereignisse(int kug_nr) {
  int i = 1;
  while(i <= anz_st) {
    if(st_ereig[anz_st].eind1==kug_nr || st_ereig[anz_st].eind2==kug_nr) {
      st_ereig[anz_st].t = unendl;
      anz_st--;
    } else if(st_ereig[i].eind1 == kug_nr || st_ereig[i].eind2 == kug_nr) {
      stosszeit a = st_ereig[anz_st];
      register int t = i;
      do {
        register int t2 = t >> 1;
      if(st_ereig[t2].t <= a.t) break;
        st_ereig[t] = st_ereig[t2];   // kleiner als die Eltern => nachrcken
        t = t2;
      } while(1);
      do {
        register int t2 = t << 1;
        if(st_ereig[t2 + 1].t < st_ereig[t2].t)
          t2++;     // rechtes Kind kleiner
      if(st_ereig[t2].t >= a.t) break;
        st_ereig[t] = st_ereig[t2];
        t = t2;
      } while(1);
      st_ereig[t] = a;
      st_ereig[anz_st].t = unendl;  // gengend grosser Wert als Stopper
      anz_st--;
    } else
      i++;
  }
}

// Nachfhren der Kugelbewegung
void kugel::bewege(float t) {
  register float dt = t - t0,
                dt2 = ldexp(dt * dt, -1);
  p.x += v.x * dt + g.x * dt2;
  p.y += v.y * dt + g.y * dt2;
  v.x += g.x * dt;
  v.y += g.y * dt;
  t0 = t;
}

void kugel::umlenkung_bande(const bandenwand &bw, float t) {
  bewege(t);
  register float vx = v.x,
                 vy = v.y;
  v.x = vx * bw.e2.x + vy * bw.e2.y; // Berechnung der
  v.y = vx * bw.e2.y - vy * bw.e2.x; // Abprallreflektion
  if(ton_ein) {
    if(sound_stat != 0) // erst krzlich ein Stoss, der immer noch ertnt?
      loesche_ereignisse(0);  // zugehriges Tonabstellereignis beseitigen
    ton_aus_sperre = 1;
    if(sound_stat != 2) {
      sound(440);
      sound_stat = 2;
    }
    stosszeit a = {t + 0.04, 0, 0};  // 40 ms lang dauert der "Ttsch"
    stoss_einfuegen(a);  // Tonabstellung als Ereignis einfgen
  }
}

// Stattfindender Stoss anhand des Impulssatzes berechnen
void kugel::umlenkung_stoss (kugel &k2, float t) {
  bewege(t);
  k2.bewege(t);
  float dx = k2.p.x - p.x,
        dy = k2.p.y - p.y,
        betr = sqrt(dx * dx + dy * dy),
        ex = dx / betr,
        ey = dy / betr,
        vu1 = v.x * ex + v.y * ey,
        vv1 = v.x * ey - v.y * ex,
        vu2 = k2.v.x * ex + k2.v.y * ey,
        vv2 = k2.v.x * ey - k2.v.y * ex,
        sm = m + k2.m,
        dm = m - k2.m,
        vu1a = (ldexp(k2.m * vu2, 1) + dm * vu1) / sm, // Anwendung Energie-
        vu2a = (ldexp(m * vu1, 1) - dm * vu2) / sm;    // und Impulssatz
  v.x = vu1a * ex + vv1 * ey;
  v.y = vu1a * ey - vv1 * ex;
  k2.v.x = vu2a * ex + vv2 * ey;
  k2.v.y = vu2a * ey - vv2 * ex;
  if(ton_ein) {
    if(sound_stat != 0) // erst krzlich ein Stoss, der immer noch ertnt?
      loesche_ereignisse (0);  // zugehriges Tonabstellereignis beseitigen
    ton_aus_sperre = 1;
    if(sound_stat != 1) {
      sound(880);
      sound_stat = 1;
    }
    stosszeit a = {t + 0.05, 0, 0};  // 50 ms lang dauert der "Ttsch"
    stoss_einfuegen(a);  // Tonabstellung als Ereignis einfgen
  }
}

// Einzelne Bandenwand prfen, wann, wenn berhaupt, ein Abprall stattfindet
int kugel::wann_bande (const bandenwand &bw, float &t) {
  float vu = v.x * bw.e.y - v.y * bw.e.x,
        pu = p.x * bw.e.y - p.y * bw.e.x;
  if(bw.g != 0.0) {
    float u = vu / bw.g,
          diskr = u * u + ldexp(bw.a - pu - r, 1) / bw.g;
    if(diskr <= 0.0) 
      return 0; // Spezialfall Streifen noch nicht ausreichend!
    if(bw.g > 0.0) {
      t = t0 - u + sqrt(diskr);
      return 1;
    }
    if(vu <= 0.0)
      return 0;
    t = t0 - u - sqrt(diskr);
    return 1;
  } else {
    // Spezialfall: Keine Beschleunigung in Querrichtung
    if(vu <= 0.0)  // muss in Richtung Bandenwand zeigen!
      return 0;
    t = t0 + (bw.a - r - pu) / vu;
    return 1;
  }
}

int kugel::wann_bande_rundum (const bandenwand b[], float &zp_min) {
  int bnr;
  zp_min = unendl;    // irgend etwas >86400 (Maximum von TIMER)
  for(int i=0; i < n_waende; i++) {
    float zp;
    if(wann_bande(b[i], zp) && zp < zp_min) {
      zp_min = zp;
      bnr = i;
    }
  }
  return bnr;
}

// Berechnung, wann ein evtl. Stoss zwischen zwei Kugeln stattfindet
int kugel::wann_stoss (const kugel &k2, float &t) {
  float dt = k2.t0 - t0,
        ux = g.x * dt,
        uy = g.y * dt,
        dvx = v.x - k2.v.x + ux,
        dvy = v.y - k2.v.y + uy;
  if(dvx == 0.0 && dvy == 0.0)
    return 0;   // genau gleiche Geschwindigkeit => kein Stoss!
  dt = ldexp(k2.t0 + t0, -1);
  float cx = p.x - k2.p.x - v.x * t0 + k2.v.x * k2.t0 - ux * dt,
        cy = p.y - k2.p.y - v.y * t0 + k2.v.y * k2.t0 - uy * dt,
        a = r + k2.r,
        u = dvx * dvx + dvy * dvy,
        diskr = a * a * u - quad(dvx * cy - dvy * cx);
  if(diskr <= 0.0)
    return 0;
  register float h = -dvx * cx - dvy * cy;
  if(h <= u * (t0 > k2.t0 ? t0 : k2.t0))  // frher stattgefundene Stsse
    return 0;  // knnen nicht mehr stattfinden!
  t = (h - sqrt(diskr)) / u;
  return 1;
}

void kugel::zeichne() {
  float dt = akt_zeit - t0,
        dt2 = ldexp(dt * dt, -1),
        mx = (p.x + dt * v.x + dt2 * g.x - x_ba) * fx,
        my = (y_ba - p.y - dt * v.y - dt2 * g.y) * fy;
  int y1 = int(mx - ra_x),
      y2 = int(mx + ra_x);
  if(y1 < x_min1)
    x_min1 = y1;
  if(y2 > x_max1)
    x_max1 = y2;
  register float yo = my - ra_y,
                 yu = my + ra_y;
  y1 = int(yo);
  y2 = int(yu);
  if(y1 < y_min1)
    y_min1 = y1;
  if(y2 > y_max1)
    y_max1 = y2;
  for(register int y=y2; y > y1; y--) {
    register float yh = float(y),
                    h = sqrt((yh - yo) * (yu - yh)) * kf;
    register int x1 = int(mx - h),
                 x2 = int(mx + h);
    if(x2 > x1)
      GrHLine(x1, x2 - 1, y, f);
  }
}

// Banden mit Przisionsalgorithmus aufzeichnen
void zeichne_bande() {
  int i = ws,
     y1 = int(floor((y_ba - bande[i].b.y) * fy));
  
  GrSetContext(leere_bande);  // als leere Bande zeichnen
  GrClearContext(0);

  // Oberer Teil
  if(y1 > -1)
    GrFilledBox(0, 0, x_aufl - 1, y1, 15);

  while(i != we) {
    float d = (bande[i].b.y - bande[(i + 1) % n_waende].b.y) * fy;
    register float a1, b1;
    if(d != 0.0) {
      a1 = (bande[(i + 1) % n_waende].b.x - bande[i].b.x) * fx / d;
      b1 = (bande[i].b.x - x_ba) * fx - a1 * (y_ba - bande[i].b.y) * fy;
    }
    i = (i + 1) % n_waende;
    int y2 = int(floor((y_ba - bande[i].b.y) * fy));
    for(register int y = y2; y > y1; y--) {
      register int x = int(float(y) * a1 + b1);
      if(x > 0)
        GrHLine(0, x - 1, y, 15);
    }
    y1 = y2;
  }

  if(y1 < y_aufl - 1)
    GrFilledBox(0, y1 + 1, x_aufl - 1, y_aufl - 1, 15);

  while(i != ws) {
    float d = (bande[i].b.y - bande[(i + 1) % n_waende].b.y) * fy;
    register float a1, b1;
    if(d != 0.0) {
      a1 = (bande[(i + 1) % n_waende].b.x - bande[i].b.x) * fx / d;
      b1 = (bande[i].b.x - x_ba) * fx - a1 * (y_ba - bande[i].b.y) * fy;
    }
    i = (i + 1) % n_waende;
    int y2 = int(floor((y_ba - bande[i].b.y) * fy));
    for(register int y = y1; y > y2; y--) {
      register int x = int(float(y) * a1 + b1);
      if(x < x_aufl)
        GrHLine(x, x_aufl - 1, y, 15);
    }
    y1 = y2;
  }
}

void main() {
  // Vorgabewerte vergeben

  for(int i=0; i < n_kugeln_max; i++) {
    b_kugel[i].t0 = akt_zeit;
    b_kugel[i].p.x = 0.0;
    b_kugel[i].p.y = 0.0;
    b_kugel[i].v.x = 0.0;
    b_kugel[i].v.y = 0.0;
    b_kugel[i].m = 100.0;
    b_kugel[i].r = 10.0;
    b_kugel[i].f = 10;
  }

  float x_min,            // linke Ausdehnung
        x_max,            // rechte Ausdehnung
        y_min,            // untere Ausdehnung
        y_max;            // obere Ausdehnung

  int k_nr = 0;     // Zeiger, welche Kugel gerade verarbeitet werden muss

  char d_name[80],  // Dateiname
       zeile[120];  // eingelesene Zeile
  FILE *d_handle;   // Datei-Handle

  // Dateiname vom Benutzer verlangen

  printf("\033[0m\033[2J\033[2;1HSimulation eines vllig reibungsfreien "
                "Billardtisches\n(c) 1996, 1997 by Andreas Meile\n\n"

         "Name einer Textdatei mit Startszenario? ");
  gets(d_name);

  // Datei einlesen und Werte fllen

  d_handle = fopen(d_name, "r");
  if(d_handle == NULL) {
    printf("Fehler beim ffnen der Datei\n");
    exit(1);
  }

  char *ende = fgets(zeile, 120, d_handle);

  while(ende != NULL) {
    char *w = zeile;
    while(*w != '%' && *w != '\0')
      w++;
    *w = '\0';      // Kommentar abschneiden
    w = zeile;
    while(*w != '=' && *w != '\0') {
      if(*w >= 'A' && *w <= 'Z')
        *w |= 0x20;
      w++;
    }
    if(*w == '=') {  // Kommandozeile
      *w = '\0';  // um Stringvergleich zu ermglichen
      w++;
      if(!strcmp(zeile, "bec")) {         // Bandenecke
        if(n_waende < n_waende_max) {     // Nur, falls noch Platz
          sscanf(w, "%f%f", &bande[n_waende].b.x, &bande[n_waende].b.y);
          if(n_waende == 0) {
            x_min = bande[n_waende].b.x;
            x_max = bande[n_waende].b.x;
            y_min = bande[n_waende].b.y;
            y_max = bande[n_waende].b.y;
            ws = n_waende;
            we = n_waende;
          } else {
            if(bande[n_waende].b.x < x_min)
              x_min = bande[n_waende].b.x;
            if(bande[n_waende].b.x > x_max)
              x_max = bande[n_waende].b.x;
            if(bande[n_waende].b.y < y_min) {
              y_min = bande[n_waende].b.y;
              we = n_waende;
            }
            if(bande[n_waende].b.y > y_max) {
              y_max = bande[n_waende].b.y;
              ws = n_waende;
            }
          }
          n_waende++;
        }
      } else if(!strcmp(zeile, "kug")) {    // Weitere kugel
        int i;
        sscanf(w, "%d", &i);
        if(i >= 1 && i <= n_kugeln_max) { // Muss im gltigen Bereich liegen
          k_nr = i - 1;
          if(i > n_kugeln)
            n_kugeln = i;
        }
      } else if(!strcmp(zeile, "pos"))    // Festlegung der Position
        sscanf(w, "%f%f", &b_kugel[k_nr].p.x, &b_kugel[k_nr].p.y);
      else if(!strcmp(zeile, "vel"))      // Geschwindigkeit
        sscanf(w, "%f%f", &b_kugel[k_nr].v.x, &b_kugel[k_nr].v.y);
      else if(!strcmp(zeile, "mas"))      // Masse
        sscanf(w, "%f", &b_kugel[k_nr].m);
      else if(!strcmp(zeile, "rad"))      // Radius
        sscanf(w, "%f", &b_kugel[k_nr].r);
      else if(!strcmp(zeile, "dic")) {    // Festlegung der Masse ber die
        float dichte;                     // Dichte
        sscanf(w, "%f", &dichte);
        b_kugel[k_nr].m = ldexp(pow(b_kugel[k_nr].r, 3), 2) * pi / 3.0
        * dichte;
      } else if(!strcmp(zeile, "far"))    // Farbe festlegen
        sscanf(w, "%d", &b_kugel[k_nr].f);
      else if(!strcmp(zeile, "ton"))      // Ton
        ton_ein = *w == 'E' || *w == 'e';
      else if(!strcmp(zeile, "vfa"))      // Verkleinerungsfaktor zur Steu-
        sscanf(w, "%f", &verkl_fakt);     // erung der Grsse am Bildschirm
      else if(!strcmp(zeile, "lfa"))      // Verlangsamungsfaktor zur
        sscanf(w, "%f", &verla_fakt);     // Anpassung an langsamere PC's
      else if(!strcmp(zeile, "bes"))      // Schwerkraft-Gravitations-
        sscanf(w, "%f%f", &g.x, &g.y);    // konstante g
      else if(!strcmp(zeile, "nei")) {    // Schwerkraft als Tischneigung mit
        float n_wink, g_betr, r_wink;     // Richtung definieren
        sscanf(w, "%f%f%f", &g_betr, &n_wink, &r_wink);
        n_wink *= (pi / 180.0);
        r_wink *= (pi / 180.0);
        g_betr *= sin(n_wink);
        g.x = g_betr * cos(r_wink);
        g.y = g_betr * sin(r_wink);
      }
    }
    ende = fgets(zeile, 120, d_handle);
  }
  fclose(d_handle);

  // Grafikmodus aktivieren
  printf("Auswahl der Grafikauflsung\n\n"
  "Wichtiger Hinweis: ltere Grafiktreiber knnen Sie nur ber die\n"
  "Treiberkonstante ansteuern, ndern Sie diese daher nur ab, falls auf\n"
  "Ihrem System die Direkt-Videomoduswahl nicht korrekt luft.\n\n");
  lies_interv_ganze_zahl("Grafiktreiberkonstante Ihres Systems", "/",
  GR_320_200_graphics, GR_width_height_color_graphics, &treiber_konst);
  
  lies_positive_zahl("Horizontale Auflsung (Anzahl Bildpunkte)", "/",
  &x_aufl);
  y_aufl = int(float(x_aufl) * y_x_verh + .9); // Addition von 0.9
  // nur, um Vorgabewert 0 zu vermeiden
  lies_positive_zahl("Vertikale Auflsung (Anzahl Bildpunkte)", "/",
  &y_aufl);
  lies_positive_zahl("Anzahl Farben (log2(Eingabe) ganzzahlig)", "/",
  &a_far);

  // Alter Videomodus sichern
  alter_videomodus = GrCurrentMode(); // alter Videomodus sichern

  GrSetMode(treiber_konst, x_aufl, y_aufl, a_far);
  // Effektive Werte auslesen
  x_aufl = GrScreenX();
  y_aufl = GrScreenY();
  a_far = GrNumColors();

  x_max1 = x_aufl - 1;
  y_max1 = y_aufl - 1;

  // Grssendaten fr Bildschirmdarstellung vorbereiten
  {
    float xm = ldexp(x_max + x_min, -1), // Mittelpunkt am Bildschirm
          ym = ldexp(y_max + y_min, -1),
          xb = (x_max - x_min) * verkl_fakt,
          yb = (y_max - y_min) * verkl_fakt;
    // Proportionen anpassen
    if(xb * y_x_verh > yb)
      yb = xb * y_x_verh;
    else
      xb = yb / y_x_verh;
    // Parameter fr Umrechnung
    fx = float(x_aufl) / xb;
    fy = float(y_aufl) / yb;
    kf = fx / fy;
    x_ba = xm - ldexp(xb, -1) - 0.5 / fx;
    y_ba = ym + ldexp(yb, -1) - 0.5 / fy;
  }

  // Umgerechnete Radienwerte fr Kreiszeichenvorgang vorberechnen sowie
  // Tempoanpassung

  g.x /= quad(verla_fakt);
  g.y /= quad(verla_fakt);
  for(i=0; i<n_kugeln; i++) {
    b_kugel[i].v.x /= verla_fakt;
    b_kugel[i].v.y /= verla_fakt;
    b_kugel[i].ra_x = b_kugel[i].r * fx;
    b_kugel[i].ra_y = b_kugel[i].r * fy;
  }

  // Banden umrechnen
  for(i=0; i<n_waende; i++) {
    float dx = bande[(i + 1) % n_waende].b.x - bande[i].b.x,
          dy = bande[(i + 1) % n_waende].b.y - bande[i].b.y,
          betr = sqrt(quad(dx) + quad(dy));
    bande[i].e.x = dx / betr;
    bande[i].e.y = dy / betr;
    bande[i].e2.x = quad(bande[i].e.x) - quad(bande[i].e.y);
    bande[i].e2.y = ldexp(bande[i].e.x * bande[i].e.y, 1);
    bande[i].a = bande[i].b.x * bande[i].e.y - bande[i].b.y * bande[i].e.x;
    bande[i].g = g.x * bande[i].e.y - g.y * bande[i].e.x;
  }

  // Grafik-Contexts anlegen
  leere_bande = GrCreateContext(x_aufl, y_aufl, NULL, NULL);
  szene = GrCreateContext(x_aufl, y_aufl, NULL, NULL);
  zeichne_bande();

  // Queue vorbereiten
  st_ereig[0].t = -unendl;  // gengend kleiner Wert als Elternsuchstopper
  for(i = 1; i < n_st_ereig_max; i++)
    st_ereig[i].t = unendl;   // gengend grosser Wert als Kindsuchstopper
  anz_st = 0;

  // Stossereigniszeiten berechnen und in die Halde einfgen

  for(i = 0; i < n_kugeln; i++) {
    sto.eind1 = ~b_kugel[i].wann_bande_rundum (bande, sto.t);
    sto.eind2 = i + 1;
    stoss_einfuegen(sto);
    for(int j = i + 1; j < n_kugeln; j++) {
      if(b_kugel[i].wann_stoss(b_kugel[j], sto.t)) {
        sto.eind1 = j + 1;
        stoss_einfuegen(sto);
      }
    }
  }

  const float tk = float(clock()) / float(CLOCKS_PER_SEC) - akt_zeit;
  // Zeitverschiebungskonstante, damit mglichst nahe bei
  // 0 gerechnet wird (Rundungsfehler.0)

  // ti[] : Zeitgltter, um den groben Takt von nur 18.2 Clocks/Sek der
  // Systemuhr auszugltten

  const int n_glaettungen = 8;   // Anzahl bercksichtigte Glttzeiten
  const int n_glaett_1 = n_glaettungen - 1;
  const float fakt = float(n_glaettungen);
  const float quot = float(n_glaettungen * (n_glaettungen + 1) >> 1);

  int sc = 0;

  float ti[n_glaettungen],
        ts = 0.0;

  for(i=0; i<n_glaettungen; i++) {
    ti[i] = float(clock()) / float(CLOCKS_PER_SEC) - tk;
    ts += ti[i];
  }
  akt_zeit = float(clock()) / float(CLOCKS_PER_SEC) - tk;

  while(!kbhit()) {
    float t1 = float(clock()) / float(CLOCKS_PER_SEC) - tk;
    akt_zeit += (fakt * t1 - ts) / quot;
    ts += t1 - ti[sc];
    ti[sc] = t1;
    sc++;
    sc &= n_glaett_1;

    if(ton_aus_sperre == 2) {
      nosound();   // Ton verzgert ausschalten, um Knacken zu verhindern
      sound_stat = 0;
      ton_aus_sperre = 0;
    }

    while(st_ereig[1].t < akt_zeit) {
      int e1 = st_ereig[1].eind1,
          e2 = st_ereig[1].eind2;
      if(e1 < 0) {  // Bandenstoss
        b_kugel[e2 - 1].umlenkung_bande(bande[~e1], st_ereig[1].t);
        loesche_ereignisse(e2);
        sto.eind1 = ~b_kugel[e2 - 1].wann_bande_rundum(bande, sto.t);
        sto.eind2 = e2;
        stoss_einfuegen(sto);
        for(i = 1; i <= n_kugeln; i++)
          if(i != e2 && b_kugel[e2 - 1].wann_stoss(b_kugel[i - 1], sto.t)) {
            sto.eind1 = i;
            stoss_einfuegen(sto);
          }
      } else if(e1 == 0) { // Nur C++: Ton abstellen, da nicht im
        if(ton_aus_sperre) {
          ton_aus_sperre = 2;
        } else {
          nosound();         // Hintergrund wie bei QuickBASIC
          sound_stat = 0;
        }
        loesche_ereignisse(0);
      } else {  // Kugelstoss
        b_kugel[e1 - 1].umlenkung_stoss(b_kugel[e2 - 1], st_ereig[1].t);
        loesche_ereignisse(e1);
        loesche_ereignisse(e2);
        i = e1;
        do {
          sto.eind1 = ~b_kugel[i - 1].wann_bande_rundum(bande, sto.t);
          sto.eind2 = i;
          stoss_einfuegen(sto);
          for(int j=1; j <= n_kugeln; j++)
            if(j != e1 && j != e2 && b_kugel[i - 1].wann_stoss(b_kugel[j
            - 1], sto.t)) {
              sto.eind1 = j;
              stoss_einfuegen(sto);
            }
          i = i == e1 ? e2 : 0;
        } while(i != 0);
      }
    }
    if(ton_aus_sperre == 1)
      ton_aus_sperre = 0;

    // Letzter Zeichenvorgang lschen
    GrSetContext(szene); // bleibt whrend der ganzen Animation
    GrBitBlt(szene, x_min1, y_min1, leere_bande, x_min1, y_min1, x_max1,
    y_max1, GrWRITE);

    // Werte sichern
    int x_min1_alt = x_min1,     // linke Begrenzung um Kugeln
        x_max1_alt = x_max1,     // rechte Begrenzung um Kugeln
        y_min1_alt = y_min1,     // obere Begrenzung um Kugeln
        y_max1_alt = y_max1;     // untere Begrenzung um Kugeln

    x_min1 = x_aufl;
    x_max1 = 0;
    y_min1 = y_aufl;
    y_max1 = -1;
    for(i=0; i<n_kugeln; i++)
      b_kugel[i].zeichne();

    y_min1++;    // wegen Beginn bei -1
    x_max1--;    // wegen Ende bei x_aufl;

    // Ausgabe auf Bildschirm: nur gerade vernderter Ausschnitt
    // aktualisieren => maximale Geschwindigkeitsoptimierung
    GrSetContext(NULL);
    GrBitBlt(NULL, x_min1 < x_min1_alt ? x_min1 : x_min1_alt,
                   y_min1 < y_min1_alt ? y_min1 : y_min1_alt,
            szene, x_min1 < x_min1_alt ? x_min1 : x_min1_alt,
                   y_min1 < y_min1_alt ? y_min1 : y_min1_alt,
                   x_max1 > x_max1_alt ? x_max1 : x_max1_alt,
                   y_max1 > y_max1_alt ? y_max1 : y_max1_alt, GrWRITE);
  }

  // Tastendruck "vernichten"
  if(getch() == '\0')
    getch();
  // Gerade noch ein Ton aktiv?
  if(sound_stat != 0)
    nosound();
  GrDestroyContext(leere_bande);
  GrDestroyContext(szene);     // wieder freigeben
  GrSetMode(alter_videomodus); // wieder in den alten Videomodus zurck

  printf("\033[0m\033[2;1HSimulation beendet.\n");
}
